(******************************************************************************
 *                                                                            *
 * Package for automatized Mellin-Barnes techniques                           *
 *                                                                            *
 * M. Czakon, 10 Oct 05                                                       *
 *                                                                            *
 * Change log:                                                                *
 *                                                                            *
 *  4 Dec 05 - added numeric interface consisting of MBintegrate,             *
 *             MBshiftContours, MBfortranForm, MBmap and MBnorm               *
 *                                                                            *
 ******************************************************************************)

Print["MB 1.1"];
Print["by Michal Czakon"];
Print["more info in hep-ph/0511200"]; 
Print["last modified 24 Jan 06"];

BeginPackage["MB`"]

(******************************************************************************
 *                                                                            *
 * Public                                                                     *
 *                                                                            *
 ******************************************************************************)

MBrules::usage = "MBrules[integrand, constraints, fixedVars] determines real
parts of fixed and integration variables such that the arguments of all Gamma
and PolyGamma functions in the integrand be positive. It is only necessary to 
specify the fixed variables, fixedVars, since the integration variables are
determined automatically. The user can also specify further constraints.\n\n
MBrules[integrand, limit, constraints, fixedVars] determines the contours such
that during analytic continuation no contour starts or ends on a pole."

MBoptimizedRules::usage = "MBoptimizedRules[integrand, limit, constraints,
fixedVars] determines optimized real parts of fixed and integration variables,
which generate the smallest number of residues. In case of a large number of
residues, one can set a limit to the number of levels, which are optimized with
the Level option. For a detailed description see MBrules and MBcontinue."

MBcorrectContours::usage = "MBcorrectContours[{fixedVarRules, intVarRules},
shift] shifts all the intVarRules by inverses of primes starting at
Prime[shift]. This can be used to avoid poles running into contours."

MBresidues::usage = "MBresidues[integrand, limit, {fixedVarRules, intVarRules}]
generates a list of residues obtained by continuing the integral given by the
integrand and the vertical integration contours in the complex plane whose real
parts are specified in intVarRules. The continuation is done to the point given
in limit. The starting point should be given in fixedVarRules. The result
contains either MBint objects where the limit has been attained, or MBitc
objects, which should be further continued. If one of the integration contours
starts or ends on a pole the process is stopped and the user has to give other
contours.\n\n
If the user is only interested in the number of integrals in the final
result, the option Skeleton -> True should be set. It might happen that there
are cancellations of the singular behaviour between the numerator and the
denominator, in which case this number will be overestimated. With the option
set, Gamma and PolyGamma functions are replaced by MBgam. With Residues -> True
two lists will be created. The first contains the details of the residues in
MBres[sign, var, val], where sign is the sign of the residue, which is taken in
var at the point val, and the second is as before."

MBcontinue::usage = "MBcontinue[integrand, limit, {fixedVarRules, intVarRules}]
analytically continues a MB integral (for a detailed description see
MBresidues). The iterative process uses at each step MBresidues and can be
stopped at a given point with the option Level."

MBmerge::usage = "MBmerge[integrals] merges integrals with the same integration
contours."

MBpreselect::usage = "MBpreselect[integrals, {x, x0, n}] rejects integrals which
vanish in the specified order n of expansion in x around x0."

MBexpand::usage = "MBexpand[integrals, norm, {x, x0, n}] expands integrals in
the specified order n of expansion in x around x0. A normalization factor norm
is applied to every integrand."

MBshiftContours::usage = "MBshiftContours[integrals] shifts the contours such
that the arguments of the Gamma and PolyGamma functions are as far away from the
poles as possible. It is assumed that the arguments contain only integration
variables, which is the case after expansion in all fixed parameters. This
function is used internally by MBintegrate."

MBfortranForm::usage = "MBfortranForm[stream, x, expression] writes an
expression of the form x = expression to the stream in Fortran form. The output
of FortranForm is corrected to have the correct syntax of double precision
constants and fit into 72 columns as required by the Fortran standard. This
function is used internally by MBintegrate."

MBmap::usage = "MBmap[z_, x_] is a mapping transformation from the range
[z - I+Infinity, z + I*Infinity], to [0, 1], as in\n\n
1/(2*Pi*I)*Integrate[f[z + I*x], {x, -Infinity, Infinity}] =
1/(2*Pi)*Integrate[MBnorm[x]*f[MBmap[z, x]], {x, 0, 1}]\n\n
This transformation is used internally by MBintegrate."

MBnorm::usage = "MBnorm[x_] for a detailed description see MBmap."

MBintegrate::usage = "MBintegrate[integrals, kinematics] numerically integrates
a list of integrals expanded in some variable with the values of the parameters
as given on the kinematics list. The input should be as produced by MBexpand.
Multiple integrals are evaluated in Fortran with the help of the CERN library
implementation of the Gamma and PolyGamma functions and of the Cuhre and Vegas
integration routines from CUBA. The libraries libmathlib.a, libkernlib.a and
libcuba.a should be available, the compilation is performed with f77. The output
is a list containing the value of the integral and the error estimates on the
real and imaginary parts. Optionally the contributions of all the numerical
integrations can be given, if option Debug is set to True. In this case, the
numerical results are contained in MBval[value, error, probability, part],
where probability is the probability that the error is underestimated, and part
is the part number of the integral.\n\n
Option description:\n\n
NamePrefix - by default the Fortran programs are called MBpart1x0, etc. where
the last number is the power of the expansion variable x. with this option one
can change MB to become NamePrefix.\n\n
PrecisionGoal, AccuracyGoal, MaxPoints, MaxRecursion - numerical integration
options as in the NIntegrate function.\n\n
MaxCuhreDim - dimension threshold, above which Vegas will be used for the
evaluation of the integrals.\n\n
Complex - by default, only the real part of the integrals is evaluated, with
this option set to True, the imaginary part will also be given.\n\n
FixedContours - contours will not be shifted if this option is set to True\n\n
NoHigherDimensional - by default, the complete integration is performed within
MBintegrate, however with this option, 1-dimensional integrals are evaluated and
the Fortran programs are prepared, but not run. This may be used to run them in
parallel for example.\n\n
Debug - with this option set to True, the Fortran programs are kept after
evaluation and the value of every integral given in the output as described
above."

barnes1::usage = "barnes1[z] applies the first Barnes lemma in the variable z."

barnes2::usage = "barnes2[z] applies the second Barnes lemma in the variable z."

Barnes1::usage = "Barnes1[MBint[integrand, {fixedVarRules, intVarRules}], z]
applies the first Barnes lemma to an integral, MBint, taking into account the
integration contours."

Barnes2::usage = "Barnes2[MBint[integrand, {fixedVarRules, intVarRules}], z]
applies the second Barnes lemma to an integral, MBint, taking into account the
integration contours."

FUPolynomials::usage = "FUPolynomials[integrand, momenta, kinematics] gives the
F and U polynomials, as well as the M and Q matrices for a Feynman integrand
given by a product of propagators\n\n
DS[k, m, n] = 1/(k^2-m^2)^n\n\n
The notation is the same as in\n\n
T. Binoth and G. Heinrich, Nucl. Phys. B585 (2000) 741."

a0::usage = "a0[ms, l] vacuum graph with one massive line to power l. a0[ms, l1,
l2] vacuum graph with one massive (power l1) and one massless line (power l2)."

b0::usage = "b0[ps, l1, l2] fully massless b0 function."

b0OS::usage = "b0OS[ps, l1, l2] on-shell b0 function. the power of the massive
line is l2."

c0::usage = "c0[ps, l1, l2, l3] fully massless c0 function with two massless
external lines. the power of the internal line that connects to both of them is
l3."

MBres::usage = "MBres[sign, var, val] denotes a residue, which is taken in var
at the point val, with sign."

MBint::usage = "MBint[integrand, {fixedVarRules, intVarRules}] is a
Mellin-Barnes integral specified by the integrand and the values of the real
parts of the fixed, fixedVarRules, and integration, intVarRules, variables."

MBitc::usage = "MBitc[integrand, limit, {fixedVarRules, intVarRules}] is a
Mellin-Barnes integral specified by the integrand and the values of the real
parts of the fixed, fixedVarRules, and integration, intVarRules, variables. The
integral still requires analytic continuation to the limit."

MBgam::usage = "MBgam[z] is a replacement for Gamma and PolyGamma functions used
by MBresidues, when the Skeleton option is set to True."

MBval::usage = "MBval[value, error, probability, part] is the value of a
numerical integral, generated with MBintegrate. part is the index of the
integral on the list of the integrals to evaluate."

ep::usage = "dimensional regularization parameter for the 1-loop integrals"

Options[MBoptimizedRules] = {
  Level -> Infinity
};

Options[MBresidues] = {
  Skeleton -> False,
  Residues -> False,
  Verbose -> True
};

Options[MBcontinue] = {
  Level -> Infinity,
  Skeleton -> False,
  Residues -> False,
  Verbose -> True
};

Options[MBintegrate] = {
  NamePrefix -> "MB",
  PrecisionGoal -> 4,
  AccuracyGoal -> 12,
  MaxPoints -> 1000000,
  MaxRecursion -> 1000,
  MaxCuhreDim -> 4,
  Complex -> False,
  FixedContours -> False,
  NoHigherDimensional -> False,
  Debug -> False,
  Verbose -> True
};

Begin["`Private`"]

(******************************************************************************
 *                                                                            *
 * Private                                                                    *
 *                                                                            *
 ******************************************************************************)

MBrules::norules = "no rules could be found to regulate this integral";

MBrules[integrand_, constraints_List, fixedVars_List] := 
Block[{fixedMap, args, integrationVars, inequalities, instance}, 

  fixedMap = (# -> 0 & ) /@ fixedVars;

  args = Cases[{integrand}, (Gamma|PolyGamma)[___, x_] -> x, -1];
  integrationVars = Union[Cases[args /. fixedMap, x_Symbol, -1]];
  inequalities =
    Union[Cases[{integrand},(Gamma|PolyGamma)[___,x_] :> x > 0, -1]];

  instance = FindInstance[
    Join[constraints, inequalities],
    Join[fixedVars, integrationVars]];

  If[instance === {}, 
    Message[MBrules::norules];
    Return[{}]];

  Return[Map[(# -> (# /. First[instance])) & ,
    {fixedVars, integrationVars}, {2}]]]

(******************************************************************************)

MBrules[integrand_, limit_Rule, constraints_List, fixedVars_List] := 
Block[{const = constraints, rules, residues},

  While[True,
    rules = MBrules[integrand, const, fixedVars]; 
    If[rules === {}, Break[]];

    residues = MBcontinue[integrand, limit, rules,
      Verbose -> False, Skeleton -> True, Residues -> True]; 

    If[Head[residues] =!= List, AppendTo[const, residues], Break[]]]; 

  Return[rules]]

(******************************************************************************)

MBoptimizedRules[integrand_,
                 limit_Rule,
                 constraints_List,
                 fixedVars_List,
                 options___Rule] :=

Block[{opt = ParseOptions[MBoptimizedRules, options], level,
  const = constraints, rules, residues, additional}, 

  level = Level /. opt;

  (* first pass, modify constraints to eliminate contours which start or end
     on a pole of some Gamma or PolyGamma function *)

  While[True,
    rules = MBrules[integrand, const, fixedVars]; 
    If[rules === {}, Break[]];

    residues = MBcontinue[integrand, limit, rules,
      Level -> level, Skeleton -> True, Residues -> True, Verbose -> False];

    If[Head[residues] =!= List, AppendTo[const, residues], Break[]]]; 

  If[rules === {}, Return[{}]];

  (* second pass, find rules which eliminate as many residues as possible *)

  const = constraints;
  additional = ((#1*(#2-#3) > 0 & ) @@ #1 & ) /@ residues[[1]] /. limit;

  Do[
    If[FreeQ[Take[additional, i-1], additional[[i]]],
      rules = MBrules[integrand, Append[const, additional[[i]]], fixedVars];
      If[rules =!= {}, AppendTo[const, additional[[i]]]]],
  {i, Length[additional]}];

  (* third pass, the same as the first *)

  While[True, 
    rules = MBrules[integrand, const, fixedVars]; 

    residues = MBcontinue[integrand, limit, rules,
      Level -> level, Skeleton -> True, Residues -> True, Verbose -> False];

    If[Head[residues] =!= List, AppendTo[const, residues], Break[]]];

  Return[rules]]

(******************************************************************************)
 
MBcorrectContours[{fixedVarRules_List, intVarRules_List}, shift_Integer] :=
  {fixedVarRules, Table[intVarRules[[i,1]] -> intVarRules[[i,2]] +
    1/Prime[shift + i - 1], {i, Length[intVarRules]}]}

(******************************************************************************)
 
MBresidues::contour = "contour starts and/or ends on a pole of Gamma[`1`]";

MBresidues[integrand_, 
           limit_Rule,
	   {fixedVarRules_List, intVarRules_List},
           options___Rule] :=

Block[{opt = ParseOptions[MBresidues, options],
  int, var, pos, args, integrals, residues = {},
  v0, v1, dir, intVars, constraint = False, pole, z,
  resVar, resSign, resRule, res, remRules, newVal,
  newIntegrand, j},
	
  (* with the skeleton option, differences between Gamma and PolyGamma are
     ignored, and no real residues are taken *)

  If[Skeleton /. opt,
    int = integrand /. (Gamma|PolyGamma)[___, x_] -> MBgam[x], int = integrand];

  var = First[limit]; 
  pos = Position[fixedVarRules, var][[1,1]];

  (* poles can be generated by Gamma and PolyGamma functions in the numerator,
     as long as they have a nontrivial dependence on integration variables. *)

  args = Union[Cases[{Numerator[int]},
    (MBgam|Gamma|PolyGamma)[___, x_] :> x /; !NumberQ[x /. fixedVarRules], -1]];

  integrals = Reap[
    Do[
      v0 = args[[i]] /. fixedVarRules /. intVarRules;
      v1 = args[[i]] /. limit /. fixedVarRules /. intVarRules;
      dir = Sign[v1 - v0];
      intVars = Union[Cases[{args[[i]]} /. fixedVarRules, x_Symbol, -1]];

      (* the algorithm gives undefined results,
         if contours start or end on a pole *)

      If[(IntegerQ[v0] && v0 <= 0) || (IntegerQ[v1] && v1 <= 0),
         Message[MBresidues::contour, args[[i]]];
         constraint = First[intVars] != (First[intVars] /. intVarRules);
         Break[]];

      If[dir != 0,

        (* find the nearest non-positive integer in the direction, in which the
           contour is shifted *)

        If[dir > 0, pole = Ceiling[v0], If[v0 < 0, pole = Floor[v0], pole = 0]];

        (* loop over all crossed poles (non-positive integers) *)

        For[z = pole, dir*(v1 - z) >= 0 && z <= 0, z += dir,
	  
          (* choose an integration variable, in which to take the residue.
             try to find one with a unit coefficient *)

          resVar = 0;
          For[j = Length[intVars], j >= 1, --j,
            If[Abs[Coefficient[args[[i]], intVars[[j]]]] == 1,
              resVar = intVars[[j]];
              Break[]]];
          If[resVar === 0, resVar = Last[intVars]];
          resSign = -dir*Sign[Coefficient[args[[i]], resVar]];
          resRule = Expand /@ Solve[args[[i]] == z, resVar][[1,1]]; 
          res = MBres[resSign, Sequence @@ resRule];

          (* do not take the same residue twice *)

          If[Count[residues, res] != 0, Continue[]];

          (* the new contours will not have resVar any more *)

          remRules = DeleteCases[intVarRules, _ ? (MemberQ[#, resVar] & )];

          (* after taking the residue the continued variable is at 
             a different place *)

          newVal = Solve[
            (args[[i]] /. var -> x /. fixedVarRules /. intVarRules) == z,
            x][[1,1]] /. x -> var;

          If[Verbose /. opt,
	     Print["Taking ", If[resSign > 0, "+", "-"],
		   "residue in ", resRule[[1]], " = ", resRule[[2]]]];

          newIntegrand = If[Skeleton /. opt,

            int /. resRule /. MBgam[x_] :> MBgam[Expand[x]] 
              /. MBgam[x_Integer] -> 1,

            Together[resSign*Residue[int, List @@ resRule]] /. {
              Gamma[x_] :> Gamma[Expand[x]],
              PolyGamma[n_, x_] :> PolyGamma[n, Expand[x]]}];

          If[newIntegrand === 0,
            If[Verbose /. opt, Print["...no contribution"]],
            AppendTo[residues, res];
            Sow[MBitc[newIntegrand, limit,
              {ReplacePart[fixedVarRules, newVal, pos], remRules}, options]]]]],

    {i, Length[args]}]];

  If[constraint =!= False, Return[constraint]];

  (* the original integral is always on the list *)

  integrals = Append[Flatten[integrals[[2]]],
    MBint[int, {ReplacePart[fixedVarRules, limit, pos], intVarRules}]];

  If[Residues /. opt,
     Return[{residues, integrals}],
     Return[integrals]]]

(******************************************************************************)

MBcontinue[integrand_,
           limit_Rule,
           {fixedVarRules_List, intVarRules_List},
           options___Rule] :=

Block[{opt = ParseOptions[MBcontinue, options], level,
  i = 1, pos, result, residues, integrals},

  level = Level /. opt;

  If[Verbose /. opt, Print["Level 1"]];
  result = MBresidues[integrand,limit,{fixedVarRules,intVarRules},options];
  If[Head[result] =!= List, Return[result]];

  If[Residues /. opt,

    residues = result[[1]];
    integrals = result[[2]],

    integrals = result];

  While[(++i <= level) && MemberQ[integrals, MBitc[__], -1],
    If[Verbose /. opt, Print["Level ", i]];
    pos = Position[integrals, MBitc[__], -1]; 

    Do[
      If[Verbose /. opt, Print["Integral ", pos[[j]]]];
      result = MBresidues @@ Extract[integrals, pos[[j]]];
      If[Head[result] =!= List, level = -1; Break[]];

      If[Residues /. opt,

        residues = Join[residues, result[[1]]];
        integrals = ReplacePart[integrals, result[[2]], pos[[j]]],

        integrals = ReplacePart[integrals, result, pos[[j]]]],

    {j, Length[pos]}]];

  If[Head[result] =!= List, Return[result]];

  If[Verbose /. opt, Print[Count[integrals, MBint[__], -1],
    " integral(s) found"]];

  If[Residues /. opt,
    Return[{residues, integrals}],
    Return[integrals]]]

(******************************************************************************)
 
MBmerge[integrals_] := 
Block[{ints = Cases[integrals, MBint[__], -1]},

  ints = ints //.
    {i1___, MBint[i2_, args__], i3___, MBint[i4_, args__], i5___} :>
    {i1, MBint[i2 + i4, args], i3, i5};
  ints = ints /. MBint[i_, args__] :> MBint[Simplify[i], args];
  ints = DeleteCases[ints, MBint[0, __], -1];
  Return[ints]]

(******************************************************************************)
 
MBpreselect[integrals_, {x_, x0_, n_}] :=
Block[{ints = Cases[integrals, MBint[__], -1],
  integrand, terms, simplified, selected},

  selected = Reap[
    Do[
      integrand = Expand[ints[[i,1]]];
      terms = If[Head[integrand] === Plus, List @@ integrand, {integrand}];
      Do[
        simplified = terms[[j]] /.
          (Gamma | PolyGamma)[___, z_] :> 1 /;  !NumberQ[z /. x -> 0];

        (* (1 + x) is needed to overcome a bug in MATHEMATICA,
           which does not expand if the function is independent of x *)
	  
        If[Normal[Series[simplified*(1 + x), {x, x0, n}]] === 0,
          terms = ReplacePart[terms, 0, {j}]],
      {j, Length[terms]}];
      If[Plus @@ terms =!= 0,
        Sow[ReplacePart[ints[[i]], Simplify[Plus @@ terms], {1}]]],
    {i, Length[ints]}]];

  If[selected[[2]] =!= {},
    Return[selected[[2,1]]],
    Return[{}]]]

(******************************************************************************)

MBexpand[integrals_, norm_, {x_, x0_, n_}] := 
Block[{ints = Cases[integrals, MBint[__], -1],
  expanded, terms, integrand},

  expanded = Table[
    terms = Expand[ints[[i,1]]];
    terms = If[Head[terms] === Plus, List @@ terms, {terms}];

    integrand = Sum[Series[norm*terms[[j]], {x, x0, n}], {j, Length[terms]}];
    integrand = Normal[Simplify[integrand]];

    MBint[integrand, ints[[i,2]]],
  {i, Length[ints]}];

  expanded = DeleteCases[expanded, MBint[0, __]];

  Return[expanded]]

(******************************************************************************)

MBshiftContours[integrals_List] := 
Block[{ints = Cases[integrals, MBint[__], -1], vars, args,
  const, val, val0, del, del0 = 0, inc = 1/100, contours},

  Do[
    vars = (First[#1] & ) /@ ints[[i,2,2]];

    If[Length[vars] > 0,
      args = Union[Cases[{ints[[i,1]]},
        (Gamma | PolyGamma)[___, x_] :> x, -1]];

      const = Table[val = args[[j]] /. ints[[i,2,2]];
        If[val < 0, val = Ceiling[val] - 1/2];
        If[val < 0, Abs[args[[j]] - val] <= del, args[[j]] >= 1/2 - del],
      {j, Length[args]}];

      While[(contours = FindInstance[const /. del -> del0, vars]) == {},
        del0 += inc];

      ints[[i,2,2]] = contours[[1]]],
  {i, Length[ints]}];
  Return[ints]]

(******************************************************************************)

MBfortranForm[stream_, x_, expression_] :=
Block[{var = ToString[x], expr, count, float = False},

  WriteString[stream, "      ", var, " = "];
  count = StringLength[var] + 3; 

  expr = ToString[expression, FortranForm];
  expr = StringReplace[expr, {"MBxPrivatex" -> "", " " -> ""}];
  expr = Characters[expr];

  (* adapting floating-point constants to double precision *)

  Do[
    If[expr[[i]] == ".",

      float = True,

      If[float,
        Which[
          expr[[i]] == "e",
          expr[[i]] = "d"; float = False,

          !DigitQ[expr[[i]]],
          expr[[i]] = {"d", "0", expr[[i]]}; float = False]]],

  {i, Length[expr]}];
  expr = Flatten[expr];

  (* output to fit into a page width of 72 characters *)

  Do[
    If[++count > 66, WriteString[stream, "\n     & "]; count = 2];
    WriteString[stream, expr[[i]]],
  {i, Length[expr]}];

  WriteString[stream, "\n"]]

(******************************************************************************)

MBmap[z_, x_] := z - I*Log[x/(1 - x)]
 
MBnorm[x_] := 1/(x*(1 - x))

(******************************************************************************)

MBintegrate::vars = "too many free variables `1`";

MBintegrate[integrals_List, kinematics_List, options___Rule] :=
Block[{opt = ParseOptions[MBintegrate, options],
  ints = integrals /. kinematics, intvars, allvars, x, greater, tolist,
  expr, ncomp, ints1, sum, pow, rules, ndim, integrand,
  name, exec, source, strm, val, result, error, acc},

  intvars = Union[Flatten[Cases[ints, MBint[_, {_, rules_}] :>
    (First[#] & /@ rules), -1]]];

  allvars = Union[Flatten[ints /. MBint[integrand_, _] :>
    Cases[N[integrand], _Symbol, -1]]];

  x = Complement[allvars, intvars];
  If[Length[x] == 0, x = {""}];
  If[Length[x] > 1, Message[MBintegrate::vars, x]; Return[{}], x = First[x]];

  greater[int1_, int2_] := Block[{e1, e2, n1, n2},
    e1 = Exponent[int1, x];
    e2 = Exponent[int2, x];
    If[e1 != e2, Return[e1 > e2]];
    n1 = int1 /. (_.)*MBint[_, {_, r_}] :> Length[r];
    n2 = int2 /. (_.)*MBint[_, {_, r_}] :> Length[r]; 
    Return[n1 > n2]];

  tolist[expr_] := If[Head[expr] === Plus, List @@ expr, {expr}];

  If[Complex /. opt, ncomp = 2, ncomp = 1];

  If[!FixedContours /. opt,
    If[Verbose /. opt, Print["Shifting contours..."]];
    ints = MBshiftContours[ints]];

  ints = Sort[Flatten[ints /. MBint[integrand_, rules_] :>
    tolist[Collect[integrand, x, MBint[#1, rules] & ]]], greater];

  sum = N[Plus @@ Cases[ints, _.*MBint[_, {_, {}}]] /. MBint[i_, _] -> i];

  ints1 = Cases[ints, _.*MBint[_, {_, {_}}]];
  If[!Complex /. opt,
    ints1 = ints1 /. MBint[integrand_, rules_] :> MBint[Re[integrand], rules]];

  If[Verbose /. opt, WriteString[$Output,
    "Performing ", Length[ints1], " 1-dimensional integrations"]];

  Do[
    If[Verbose /. opt, WriteString[$Output, "...", i]];
    sum += ints1[[i]] /. MBint[integrand_, {_, rules_}] :>
      NIntegrate[1/(2*Pi)*MBnorm[MBx]*integrand /. 
        rules[[1,1]] -> MBmap[rules[[1,2]], MBx], {MBx, 0, 1},
        PrecisionGoal -> (PrecisionGoal /.opt),
        AccuracyGoal -> (AccuracyGoal /. opt),
        MaxRecursion -> (MaxRecursion /. opt),
        Method -> GaussKronrod],
  {i, Length[ints1]}];

  If[!Complex /. opt, sum = Collect[sum, x, Re]];

  ints = DeleteCases[ints, _.*MBint[_, {_, {}|{_}}]];

  If[Verbose /. opt, Print["Higher-dimensional integrals"]];

  Do[
    pow = Exponent[ints[[i]], x];
    rules = ints[[i]] /. _.*MBint[_, {_, r_}] -> r;
    ndim = Length[rules];
    integrand = ints[[i]] /. _.*MBint[j_, _] -> j;
    name = StringJoin[NamePrefix /. opt, "part", ToString[i]];
    exec = StringJoin[name, ToString[x], ToString[pow]];
    source = StringJoin[exec, ".f"];
    strm = OpenWrite[source];

    If[Verbose /. opt, Print["Preparing ", exec, " (dim ", ndim, ")"]];

    WriteString[strm,
      "      implicit none\n\n",

      "      external integrand\n",
      "      integer ndim, ncomp, flags, mineval, maxeval\n",
      "      integer key, nstart, nincrease\n",
      "      double precision epsrel, epsabs\n\n",

      "      parameter (ndim=", ndim, ", ncomp=", ncomp, ", flags=0,\n",
      "     & mineval=0, maxeval=", FortranForm[MaxPoints /. opt], ",\n",
      "     & key=0, nstart=1000, nincrease=1000,\n",
      "     & epsrel=", FortranForm[10^(-PrecisionGoal) /. opt], ",\n",
      "     & epsabs=", FortranForm[10^(-AccuracyGoal) /. opt], ")\n\n",

      "      integer nregions, neval, fail\n",
      "      double precision integral(ncomp), error(ncomp), prob(ncomp)\n\n",

      "      if (ndim.gt.", MaxCuhreDim /. opt, ") then\n",
      "        call vegas(ndim, ncomp, integrand, epsrel, epsabs, flags,\n",
      "     &   mineval, maxeval, nstart, nincrease,\n",
      "     &   neval, fail, integral, error, prob)\n",
      "      else\n",
      "        call cuhre(ndim, ncomp, integrand, epsrel, epsabs, flags,\n",
      "     &   mineval, maxeval, key, nregions,\n",
      "     &   neval, fail, integral, error, prob)\n",
      "      endif\n\n",

      "      write (*,*) integral(1), error(1), prob(1)\n",
      "      if (ncomp.eq.2) then\n",
      "        write (*,*) integral(ncomp), error(ncomp), prob(ncomp)\n",
      "      endif\n\n",

      "      end\n\n"];

    WriteString[strm,
      "      subroutine integrand(ndim, MBx, ncomp, MBf)\n\n",

      "      integer ndim, ncomp\n",
      "      double precision MBx(ndim), MBf(ncomp)\n",
      "      double complex MBval, wgamma, wpsipg\n"];
    Do[WriteString[strm,
      "      double complex ", rules[[j,1]], "\n"], {j, ndim}];

    WriteString[strm, "\n"];
    Do[MBfortranForm[strm, rules[[j,1]],
      MBmap[rules[[j,2]], ToExpression[StringJoin["MBx[", ToString[j], "]"]]]],
    {j, ndim}];

    WriteString[strm, "\n"];
    MBfortranForm[strm, MBval,
      Product[MBnorm[ToExpression[StringJoin["MBx[", ToString[j], "]"]]],
      {j, ndim}]*N[(1/(2*Pi)^ndim)*integrand] /.
      {Gamma[z_] -> wgamma[z], PolyGamma[n_, z_] :> wpsipg[z, n]}];
    WriteString[strm, "\n"];

    WriteString[strm,
      "      MBf(1) = dreal(MBval)\n",
      "      if (ncomp.eq.2) MBf(ncomp) = dimag(MBval)\n\n",

      "      end\n"];

    Close[strm];

    Run["f77 -O -o", exec, source, "-L. -lmathlib -lkernlib -lcuba"],
  {i, Length[ints]}];

  If[NoHigherDimensional /. opt,
    If[Verbose /. opt,
      Print["The result does not contain higher dimensional integrals"]];
    Return[sum]];

  Do[
    pow = Exponent[ints[[i]], x];
    name = StringJoin[NamePrefix /. opt, "part", ToString[i]];
    exec = StringJoin[name, ToString[x], ToString[pow]];
  
    If[Verbose /. opt, Print["Running ", exec]];

    val = Import[StringJoin["! ./", exec], "Table"];
    val = If[ncomp == 1, Flatten[val], Table[val[[1,j]]+I*val[[2,j]], {j,3}]];
    sum += x^pow*MBval[Sequence @@ val, i];

    If[!Debug /. opt,
      DeleteFile[StringJoin[name, ToString[x], ToString[pow], ".f"]];
      DeleteFile[exec]],
  {i, Length[ints]}];

  result = sum /. MBval[i_, __] -> i;
  error = Collect[sum, x, 
    Sqrt[Plus @@ Cases[tolist[#], MBval[_, i_, __] :> {Re[i]^2, Im[i]^2}]] &];

  If[Debug /. opt,
    Return[{result,error,sum}],
    Return[{result,error}]]]

(******************************************************************************)
 
barnes1[z_] := Gamma[z + (l1_.)]*Gamma[z + (l2_.)]*Gamma[-z + (l3_.)]*
      Gamma[-z + (l4_.)] -> Gamma[l1 + l3]*Gamma[l1 + l4]*Gamma[l2 + l3]*
      (Gamma[l2 + l4]/Gamma[l1 + l2 + l3 + l4])

barnes2[z_] := Gamma[(l1_.) + z]*Gamma[(l2_.) + z]*Gamma[(l3_.) + z]*
      Gamma[(l4_.) - z]*(Gamma[(l5_.) - z]/Gamma[(l6_.) + z]) :> 
     Gamma[l1 + l4]*Gamma[l2 + l4]*Gamma[l3 + l4]*Gamma[l1 + l5]*
       Gamma[l2 + l5]*(Gamma[l3 + l5]/(Gamma[l1 + l2 + l4 + l5]*
         Gamma[l1 + l3 + l4 + l5]*Gamma[l2 + l3 + l4 + l5])) /; 
      Expand[l1 + l2 + l3 + l4 + l5 - l6] === 0

Barnes1[MBint[integrand_, {fixedVarRules_List, intVarRules_List}], z_Symbol] := 
Block[{original = MBint[integrand, {fixedVarRules, intVarRules}],
  rules = Flatten[{fixedVarRules, intVarRules}],
  x, pos, a, b, p, neg, c, d, n, xx, xx0, int, fixed, cont, ct, orig},

  pos = Cases[{Numerator[integrand]},
    Gamma[x_] :> x /; Coefficient[x, z] == +1, -1] /. z -> 0;
  If[Length[pos] == 1, pos = {First[pos], First[pos]}];
  If[Length[pos] != 2, Return[original]];
  a = First[pos];
  b = Last[pos];

  neg = Cases[{Numerator[integrand]},
    Gamma[x_] :> x /; Coefficient[x, z] == -1, -1] /. z -> 0;
  If[Length[neg] == 1, neg = {First[neg], First[neg]}];
  If[Length[neg] != 2, Return[original]];
  c = First[neg];
  d = Last[neg];

  If[MemberQ[integrand /.
    Gamma[a + z]*Gamma[b + z]*Gamma[c - z]*Gamma[d - z] -> 1, z, -1],
    Return[original]];

  p = -Min[pos /. rules];
  n = +Min[neg /. rules];
  xx0 = Max[{p - z, z - n} /. rules];

  If[xx0 < 0,
    Return[DeleteCases[original /.
      Gamma[a + z]*Gamma[b + z]*Gamma[c - z]*Gamma[d - z] ->
      Gamma[a + c]*Gamma[a + d]*Gamma[b + c]*Gamma[b + d]/Gamma[a + b + c + d],
      z -> _, -1]]];

  int = integrand /. Gamma[a + z]*Gamma[b + z]*Gamma[c - z]*Gamma[d - z] ->
    Gamma[a + xx + z]*Gamma[b + xx + z]*Gamma[c + xx - z]*Gamma[d + xx - z];

  fixed = DeleteCases[Join[fixedVarRules, intVarRules, {xx -> xx0 + 10^-10}],
    z -> _, -1];

  cont = MBmerge[MBcontinue[int, xx -> 0, {fixed, {z -> (z /. intVarRules)}},
    Verbose -> False]];

  ct = -cont[[1,1]];
  orig = integrand /. Gamma[a + z]*Gamma[b + z]*Gamma[c - z]*Gamma[d - z] ->
    Gamma[a + c + 2*xx]*Gamma[a + d + 2*xx]*
    Gamma[b + c + 2*xx]*Gamma[b + d + 2*xx]/Gamma[a + b + c + d + 4*xx];

  Return[MBint[Normal[Simplify[Series[orig + ct, {xx, 0, 0}]]],
    {fixedVarRules, DeleteCases[intVarRules, z -> _, -1]}]]]
 
Barnes2[MBint[integrand_, {fixedVarRules_List, intVarRules_List}], z_Symbol] := 
Block[{original = MBint[integrand, {fixedVarRules, intVarRules}], sign = +1,
  integ = integrand, rules = Flatten[{fixedVarRules, intVarRules}], x, i, j,
  tpos, pos = {}, a, b, c, p, tneg, neg = {}, d, e, n, 
  xx, xx0, int, fixed, cont, ct, orig},

  If[MemberQ[{Denominator[integrand]}, Gamma[x_.-z], -1],
    sign = -1;
    integ = integ /. z -> -z;
    rules = rules /. (z -> x_) -> (z -> -x)];

  tpos = Cases[{Numerator[integ]},
    Gamma[x_] :> x /; Coefficient[x, z] == +1, -1] /. z -> 0;

  Do[
    Do[AppendTo[pos, tpos[[i]]],
    {j, Exponent[integ, Gamma[tpos[[i]] + z]]}],
  {i, Length[tpos]}];

  If[Length[pos] != 3, Return[original]]; 

  a = pos[[1]];
  b = pos[[2]];
  c = pos[[3]];

  tneg = Cases[{Numerator[integ]},
    Gamma[x_] :> x /; Coefficient[x, z] == -1, -1] /. z -> 0;

  Do[
    Do[AppendTo[neg, tneg[[i]]],
    {j, Exponent[integ, Gamma[tneg[[i]] - z]]}],
  {i, Length[tneg]}];

  If[Length[neg] != 2, Return[original]]; 

  d = neg[[1]];
  e = neg[[2]];

  If[MemberQ[integ /.
    Gamma[a + z]*Gamma[b + z]*Gamma[c + z]*Gamma[d - z]*Gamma[e - z]/
    Gamma[Expand[a + b + c + d + e] + z] -> 1, z, -1],
    Return[original]];

  p = -Min[pos /. rules];
  n = +Min[neg /. rules];
  xx0 = Max[{p - z, z - n} /. rules];

  If[xx0 < 0,
    Return[DeleteCases[original /. z -> sign*z /.
      Gamma[a + z]*Gamma[b + z]*Gamma[c + z]*Gamma[d - z]*Gamma[e - z]/
      Gamma[Expand[a + b + c + d  + e] + z] ->
      Gamma[a + d]*Gamma[b + d]*Gamma[c + d]*
      Gamma[a + e]*Gamma[b + e]*Gamma[c + e]/
      (Gamma[a + b + d + e]*Gamma[a + c + d + e]*Gamma[b + c + d + e]),
      sign*z -> _, -1]]];

  int = integ /.
    Gamma[a + z]*Gamma[b + z]*Gamma[c + z]*Gamma[d - z]*Gamma[e - z]/
    Gamma[Expand[a + b + c + d  + e] + z] ->
    Gamma[a + xx + z]*Gamma[b + xx + z]*Gamma[c + xx + z]*
    Gamma[d + xx - z]*Gamma[e + xx - z]/
    Gamma[Expand[a + b + c + d  + e] + 5*xx + z];

  fixed = DeleteCases[Join[fixedVarRules, intVarRules, {xx -> xx0 + 10^-10}],
    z -> _, -1];

  cont = MBmerge[MBcontinue[int, xx -> 0,
    {fixed, {z -> (sign*z /. intVarRules)}}, Verbose -> False]];

  ct = -cont[[1,1]];
  orig = integ /.
    Gamma[a + z]*Gamma[b + z]*Gamma[c + z]*Gamma[d - z]*Gamma[e - z]/
    Gamma[Expand[a + b + c + d  + e] + z] ->
    Gamma[a + d + 2*xx]*Gamma[b + d + 2*xx]*Gamma[c + d + 2*xx]*
    Gamma[a + e + 2*xx]*Gamma[b + e + 2*xx]*Gamma[c + e + 2*xx]/
    (Gamma[a + b + d + e + 4*xx]*Gamma[a + c + d + e + 4*xx]*
    Gamma[b + c + d + e + 4*xx]);

  Return[MBint[Normal[Simplify[Series[orig + ct, {xx, 0, 0}]]],
    {fixedVarRules, DeleteCases[intVarRules, z -> _, -1]}]]]
 
(******************************************************************************)

FUPolynomials[integrand_, momenta_, kinematics_] :=
Block[{denom = List @@ integrand, int, l = Length[momenta], m, j, q, u},

  int = Sum[Global`x[i]*denom[[i]] /. Global`DS[k_, m_, _] -> k^2 - m^2,
    {i, Length[denom]}];
  m = Table[If[i != j, 1/2, 1]*
    Coefficient[int, momenta[[i]]*momenta[[j]]], {i, l}, {j, l}];
  j = int - momenta.m.momenta;
  q = Table[Expand[-1/2*Coefficient[j, momenta[[i]]]], {i, l}] ;
  j = Expand[j + 2*momenta.q] /. kinematics;
  u = Det[m];
  Return[{Simplify[u*(Expand[q.Inverse[m].q] - j) /. kinematics], u, m, q}]]

(******************************************************************************)
 
a0[ms_, l_] := Gamma[l + ep - 2]/(Gamma[l]*ms^(l + ep - 2))

a0[ms_, l1_, l2_] := Gamma[l1 + l2 + ep - 2]*(Gamma[-l2 - ep + 2]/
      (Gamma[l1]*Gamma[2 - ep]*ms^(l1 + l2 + ep - 2)))

b0[ps_, l1_, l2_] := Gamma[2 - ep - l1]*Gamma[2 - ep - l2]*
     (Gamma[l1 + l2 + ep - 2]/(Gamma[l1]*Gamma[l2]*Gamma[4 - l1 - l2 - 2*ep]*
       (-ps)^(l1 + l2 + ep - 2)))

b0OS[ps_, l1_, l2_] := Gamma[l1 + l2 + ep - 2]*(Gamma[-2*l1 - l2 - 2*ep + 4]/
      (Gamma[l2]*Gamma[-l1 - l2 - 2*ep + 4]*ps^(l1 + l2 + ep - 2)))

c0[ps_, l1_, l2_, l3_] := Gamma[-l1 - l3 - ep + 2]*Gamma[-l2 - l3 - ep + 2]*
     (Gamma[l1 + l2 + l3 + ep - 2]/(Gamma[l1]*Gamma[l2]*
       Gamma[-l1 - l2 - l3 - 2*ep + 4]*(-ps)^(l1 + l2 + l3 + ep - 2)))

(******************************************************************************
 *                                                                            *
 * Return a list, where all the options are set, either as on the default     *
 * option list, or as specified by the user.                                  *
 *                                                                            *
 ******************************************************************************)

ParseOptions[f_, options___Rule] :=
Block[{pos},
  If[Length[pos = Position[{options}, First[#] -> _, 1, 1] ] === 0, #,
    {options}[[ Sequence @@ First[pos]]]] & /@ Options[f]]

End[]

EndPackage[]
